package
{
	import mx.containers.Grid;
	public class GridState
	{
		public const width:int=10;
		public const height:int=8;
		
		public static var warmwalls:Boolean = false;
		
		private var squares:Vector.<Vector.<int>>;
		
		public function GridState()
		{
			squares = new Vector.<Vector.<int>>();
			for (var i:int=0;i<width;i++)
			{
				var row:Vector.<int> = new Vector.<int>();
				for (var j:int=0;j<height;j++)
				{
					row.push(3);
				}
				squares.push(row);
			}
		}
		
		public function IsWarm(tilekind:int):Boolean
		{
			if( warmwalls)
			{
				return tilekind==1 || tilekind==2 || tilekind==4 || tilekind==5;	
			}
			else
			{
				return tilekind==2 || tilekind==4 || tilekind==5;
			}
		}
		
		public function IsPushable(tilekind:int):Boolean
		{		
				return tilekind==2 || tilekind==4 || tilekind==5;
		}
		
		public function IsPushOnAble(tilekind:int):Boolean
		{		
			return tilekind==0;
		}
		
		public function AtPeace(x:int,y:int):Boolean
		{
			var t:int=squares[x][y];
			if (t!=2 && t!= 4 && t!=5)
				return true;
			
			var security:int = Security(x,y);
			
			if (t==2) //good
			{
				if (security==0)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			else if (t==4) // bad
			{
				if (security==2)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			else// if (t==5) //neutral
			{
				if (security==1)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
		}
		
		public function Solved():Boolean
		{
			var result:GridState = new GridState();
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					if (AtPeace(i,j)==false)
						return false;
				}
			}
			return true;
		}
		/*
			-1 	- not cat
			2	- unhappy
			1	- unstable-secure
			0	- secure
		*/
		public function Security(i:int,j:int):int
		{
			var t:int = GetTile(i,j);
			if (t!=2 && t!=4 && t!=5)
				return -1;
								
			var warm:Boolean=false;
			warm = warm || IsWarm(GetTile(i-1,j));
			warm = warm || IsWarm(GetTile(i+1,j));
			warm = warm || IsWarm(GetTile(i,j-1));
			warm = warm || IsWarm(GetTile(i,j+1));
			if (warm==false)
				return 2;
			
			for (var d:int=0;d<4;d++)
			{
				var g:GridState = PushTile(i,j,d);
				var dx:int=0;
				var dy:int=0;
				switch(d)
				{
					case 0:
						dy=-1;
						break;
					case 1:
						dy=1;
						break;
					case 2:
						dx=-1;
						break;
					case 3:
						dx=1;
						break;
				}
				if (g!=null)
				{					
					var warmneighbour:Boolean=false;
					warmneighbour = warmneighbour || IsWarm(g.GetTile(i-1+dx,j+dy));
					warmneighbour = warmneighbour || IsWarm(g.GetTile(i+1+dx,j+dy));
					warmneighbour = warmneighbour || IsWarm(g.GetTile(i+dx,j-1+dy));
					warmneighbour = warmneighbour || IsWarm(g.GetTile(i+dx,j+1+dy));
					if (warmneighbour==false)
						return 1;
					
				}
			}
				
			return 0;			
		}
		
		public function GetTile(x:int,y:int):int
		{
			if (x<0 || y<0 || x>=width || y>=height)
			{
				return -1;
			}
			return squares[x][y];
		}
		
		public function SetTile(x:int,y:int,val:int):GridState
		{
			if (x<0 || y<0 || x>=width || y>=height)
			{
				trace("tile out of bounds " + x + "," + y + " - " + val);
				return null;
			}
			
			var result:GridState = Copy();
			result.squares[x][y]=val;
			return result;
		}
		
		
		public function SetTiles(a:Array,val:int):GridState
		{
			var result:GridState = Copy();
			for (var i:int=0;i<a.length;i++)
			{
				var x:int=a[i][0];
				var y:int=a[i][1];
				
				if (x<0 || y<0 || x>=width || y>=height)
				{
					trace("tile out of bounds " + x + "," + y + " - " + val);
					return null;
				}
			
				result.squares[x][y]=val;
			}
			
			return result;
		}
		
		public function PushTile(x:int,y:int,dir:int):GridState
		{
			var tx:int;
			var ty:int;
			switch(dir)
			{
				case 0:
					tx=x;
					ty=y-1;
					break;
				case 1:
					tx=x;
					ty=y+1;
					break;
				case 2:
					tx=x-1;
					ty=y;
					break;
				case 3:
					tx=x+1;
					ty=y;
					break;
			}
			var t:int = GetTile(x,y);
			var tt:int = GetTile(tx,ty);
			if (IsPushable(t) && IsPushOnAble(tt))
			{
				var result:GridState = Copy();
				result = result.SetTile(x,y,0);
				result = result.SetTile(tx,ty,t);
				return result;
			}
			else
			{
				return null;
			}
			
		}
		
		private function Copy():GridState
		{
			var result:GridState = new GridState();
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					result.squares[i][j]=squares[i][j];
				}
			}
			return result;
		}
		
		public function Serialize():String
		{
			var result:String = "";
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					result += squares[i][j].toString()+",";
				}
			}
			return result
		}
		
		public static function Deserialize(s:String):GridState
		{
			var result:GridState = new GridState();
			var a:Array = s.split(",");
			var k:int=0;
			
			for (var i:int=0;i <result.width;i++)
			{
				for (var j:int=0;j<result.height;j++)
				{					
					result.squares[i][j]=int(a[k]);
					k++;
				}
			}
			return result
		}
		
		public function GetPieceCoords():Array
		{
			var result:Array = new Array();
			
			for (var i:int=0;i <width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					if (squares[i][j]==0 || squares[i][j]==2 || squares[i][j]==4 || squares[i][j]==5)
					{
						result.push([i,j]);
					}
				}
			}			
			
			return result;
		}
		
		public function Stratify(source:Array):Array
		{
			var maxlength:int=0;
			for (var i:int=0;i<source.length;i++)
			{
				if (maxlength<source[i].length)
				{
					maxlength=source[i].length;
				}
			}
			
			var result:Array = new Array();
			
			for (i=0;i<maxlength;i++)
			{
				var r:Array = new Array();
				for (var j:int=0;j<source.length;j++)
				{
					if (source[j].length==i+1)
					{
						r.push(source[j]);
					}
				}
				result.push(r);
			}
			return result;
		}
		
		public static function Subsets(source:Array ):Array
		{
			if (source.length<2)
			{
				return [source];
			}
			
			
			var tail:Array = source.concat();
			tail.splice(i,1);
			var tailpow:Array = Subsets(tail);
			
			var result:Array = new Array();
			var s0:Object = source[0];
			
			result.push([s0]);
			
			for (var i:int=0;i<tailpow.length;i++)
			{
				
				result.push(tailpow[i]);
				var a:Array = tailpow[i].concat();
				a.push(s0);
				
				result.push(a);
			}
			
			return result;
		}
		
		public function RemoveAll(tile:int):GridState
		{
			var result:GridState = Copy();
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					if (result.squares[i][j]==tile)
					{
						result.squares[i][j]=0;
					}
				}				
			}
			return result;
		}
		
		public function StableImmediateSubgrids():Vector.<GridState>
		{
			var result:Vector.<GridState> = new Vector.<GridState>();
			
			var emptied:GridState = RemoveAll(2).RemoveAll(4).RemoveAll(5);
			
			var piececoords:Array = emptied.GetPieceCoords();			
			
			for(var i:int=0;i<piececoords.length;i++)
			{
				var g:GridState = emptied.SetTile(piececoords[i][0],piececoords[i][1],0);
				if (g.Stable())
				{
					result.push(g);
				}
			}
			return result;			
		}		
		
		
		private static function comp(g1:GridState, g2:GridState): int
		{
			return g1.Size()-g2.Size();
		}
		
		public function AllStableSubgrids():Vector.<GridState>
		{
			var result:Vector.<GridState> = new Vector.<GridState>();
			
			var emptied:GridState = RemoveAll(2).RemoveAll(4).RemoveAll(5);
			
			var piececoordsubsets:Array = Subsets(emptied.GetPieceCoords());
			trace("subsetlength " + piececoordsubsets.length);
			
			var strata:Array = Stratify(piececoordsubsets);
			trace("strata count = " + strata.length);
			//strata.reverse();
			
			
			for (var s:int=0;s<strata.length;s++)
			{				
				var stratum:Array = strata[s];
				trace("current stratum " + s + " num entries = " + stratum.length);
				for (var i:int=0;i<stratum.length;i++)
				{
					var g:GridState = emptied.SetTiles(stratum[i],2);
					if (g.Stable())
					{
						result.push(g);
					}
				}
				
				if (result.length>0)
					break;
			}

			trace("result count " + result.length);
			return result;
		}
		
		public function Equals(other:GridState):Boolean
		{
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					if (squares[i][j]!=other.GetTile(i,j))
						return false;
				}
			}
			return true;
		}
		
		public function Size():int
		{
			var result:int=0;
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					if (squares[i][j]==2 || squares[i][j]==4 || squares[i][j]==5)
					{
						result++;
					}
				}				
			} 
			return result;
		}
		
		public function Stable():Boolean
		{
			var alterations:Vector.<GridState> = new Vector.<GridState>();
			alterations.push(this);
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					if (GetTile(i,j)==2)
					{
						for (var d:int=0;d<4;d++)
						{
							var g:GridState = PushTile(i,j,d);
							if (g!=null)
							{
								alterations.push(g);
							}
						}
					}
				}
			}
			
			for (i=0;i<alterations.length;i++)
			{
				if (alterations[i].Broken())
					return false;
			}
			return true;
		}
		
		public function Broken():Boolean
		{
			var foundone:Boolean=false;
			for (var i:int=0;i<width;i++)
			{
				for (var j:int=0;j<height;j++)
				{
					if (squares[i][j]==2)
					{
						foundone=true;
						var warmneighbour:Boolean=false;
						warmneighbour = warmneighbour || IsWarm(GetTile(i-1,j));
						warmneighbour = warmneighbour || IsWarm(GetTile(i+1,j));
						warmneighbour = warmneighbour || IsWarm(GetTile(i,j-1));
						warmneighbour = warmneighbour || IsWarm(GetTile(i,j+1));
						if (warmneighbour==false)
							return true;
					}
				}
			}
			return !foundone;
		}
	}
}